module eta_pspi

  use ani_type
  use ani_parms_cg
  use ani_fftw 

  implicit none

  contains

  ! =================================================================================
  ! Main pspi module
  subroutine etapspi(iwa, inode, sign, ws, tflds, model, refr, gat)

    character*3					:: gat
    integer                                     :: iwa, inode, sign
    complex                                     :: ws(:), tflds(:)
    complex, allocatable                        :: cube(:,:)
    real                                        :: model(:,:), refr(:,:)

    real                                        :: s, eta, dlt
    integer                                     :: nref, iref

    allocate(cube(nx,nref_max))

    do iref=1, nref_max
      cube(:,iref) = ws
    end do

    call select_ref(refr(:,1), nref)
    do iref = 1, nref
      s   = refr(iref,1) 
      eta = refr(iref,2)
      dlt = refr(iref,3)
!      write(0,*) "v=",v,"eta=",eta,"dlt",dlt
      call eta_ssf_exp(iwa, inode, cube(:,iref), tflds, model(:,1), s, eta, dlt, sign, gat)
    end do
    call interp_lloyd(cube, ws, model, refr, scales, nref)

    deallocate(cube)
  end subroutine etapspi

  ! =================================================================================
  ! One step in depth extrapolation
  subroutine eta_ssf_exp(iwa, inode, DATw, tflds, sxmodel, s, eta, dlt, sign, gat)
    implicit none
    complex                                     :: DATw(:), tflds(:)
    real                                        :: sxmodel(:)
    real                                        :: dw, dkx, w
    real                                        :: wov, wov2, kx, kx2, kz, kz2, delta_s
    integer                                     :: ikx, jkx, iwa, ix, nkx, sign, inode
    character*3					:: gat
    integer*8, allocatable                      :: planF(:)
    integer*8, allocatable                      :: planI(:)

    real     :: s, eta, dlt
    real     :: phase, time

    allocate(planF(node), planI(node))
    nkx = kxm%n
    dkx = kxm%d * 2*pi

    if (gat.eq."sou") then
      tflds=DATw
      call ani_fftexes(tflds,inode)
      DATw=tflds/sqrt(1.*kxm%n)
    else
      tflds=DATw
      call ani_fftexer(tflds,inode)
      DATw=tflds/sqrt(1.*kxm%n)
    end if

    w = rec%w%all(iwa)
    if(w.eq.0.0) w=1.0e-06
    wov=w*s
    wov2=wov*wov

    do ikx = 1, nkx
      kx = kxm%all(ikx)
      kx2 = kx * kx
      kz2 = wov2*(wov2-kx2*(1.0+2.0*eta))/(wov2-2.0*kx2*eta)*(1+2*dlt)      
      if(kz2.ge.0.0) then
        kz=sign*sqrt(kz2)
        phase=-kz*dz
        DATw(ikx)=DATw(ikx)*cmplx(cos(phase), sin(phase))
      else  ! evanescent energy
        DATw(ikx)=cmplx(0.0,0.0)
      end if       
    end do

    if (gat.eq."sou") then
      tflds=DATw
      call ani_ifftexes(tflds,inode)
      DATw=tflds/sqrt(1.*kxm%n)
    else
      tflds=DATw
      call ani_ifftexer(tflds,inode)
      DATw=tflds/sqrt(1.*kxm%n)
    end if

    do ix=1, nx
      delta_s=sxmodel(ix)-s
      time=-w*dz*delta_s*sign
      DATw(ix)=DATw(ix)*cmplx(cos(time), sin(time))
    end do
    deallocate(planF, planI) 
  end subroutine eta_ssf_exp

  ! =================================================================================
  ! Lloyd interpolation
  subroutine interp_lloyd(cube, wavefield, ani_modx, ani_refx, scale, nref)

    implicit none
    complex :: cube(:,:)
    complex :: wavefield(:)
    real :: ani_modx(:,:), ani_refx(:,:), scale(:)
    real, allocatable :: cent_modx(:), cent_refx(:)

    real    :: cmin, cmax, dc, error, error_min
    integer :: nref, lc, ipar, ix, iref

    if(nref==1)then
      do ix=1, nx
        wavefield(ix)=cube(ix, 1)
      end do
      return
    end if

    allocate(cent_modx(nx), cent_refx(nref))

    cent_modx=0.0
    cent_refx=0.0

    do ipar=1, npar_in

      cent_modx=cent_modx + ani_modx(:,ipar)*scale(ipar)
      cent_refx=cent_refx + ani_refx(:,ipar)*scale(ipar)

    end do

    cent_modx=cent_modx/npar_in
    cent_refx=cent_refx/npar_in

    do ix=1, nx

      error_min=abs(cent_modx(ix)-cent_refx(1))
      lc=1
      do iref=2, nref

        error=abs(cent_modx(ix)-cent_refx(iref))
        if(error<error_min)then
          error_min=error
          lc=iref
        end if

      end do
      wavefield(ix)=cube(ix, lc)
    end do

    deallocate(cent_modx, cent_refx)
    return
  end subroutine interp_lloyd


  ! =================================================================================
  ! Reference model selection module
  subroutine select_ref(array, ncount)
    implicit none
    integer :: iref, nref_max, ncount
    real    :: array(:)

    nref_max=size(array)

    if(array(1)<=0.0)then
      ncount=1
      return
    else
      ncount=0
      do iref=1, nref_max
        ncount=ncount+1
        if(array(iref)<=0.0)then
          ncount=iref-1
          return
        end if
      end do
    end if
    return
  end subroutine select_ref

end module eta_pspi
